// Created by: Eugeny MALTCHIKOV
// Copyright (c) 2015 OPEN CASCADE SAS
//
// This file is part of Open CASCADE Technology software library.
//
// This library is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License version 2.1 as published
// by the Free Software Foundation, with special exception defined in the file
// OCCT_LGPL_EXCEPTION.txt. Consult the file LICENSE_LGPL_21.txt included in OCCT
// distribution for complete text of the license and disclaimer of any warranty.
//
// Alternatively, this file may be used under the terms of Open CASCADE
// commercial license or contractual agreement.


#include <BOPAlgo_CellsBuilder.hxx>

#include <TopoDS_Compound.hxx>

#include <BRep_Builder.hxx>

#include <TopExp_Explorer.hxx>

#include <BOPDS_DS.hxx>

#include <BOPTools.hxx>
#include <BOPTools_AlgoTools.hxx>
#include <BOPTools_AlgoTools3D.hxx>

#include <BOPAlgo_BuilderSolid.hxx>

#include <ShapeUpgrade_UnifySameDomain.hxx>


static
  TopAbs_ShapeEnum TypeToExplore(const Standard_Integer theDim);

static
  void MakeTypedContainers(const TopoDS_Shape& theSC,
                           const TopAbs_ShapeEnum aType,
                           TopoDS_Shape& theResult);

//=======================================================================
//function : empty constructor
//purpose  : 
//=======================================================================
BOPAlgo_CellsBuilder::BOPAlgo_CellsBuilder()
:
  BOPAlgo_Builder(),
  myType(TopAbs_SHAPE),
  myIndex(100, myAllocator),
  myMaterials(100, myAllocator),
  myShapeMaterial(100, myAllocator),
  myMapModified(100, myAllocator)
{
}

//=======================================================================
//function : empty constructor
//purpose  : 
//=======================================================================
BOPAlgo_CellsBuilder::BOPAlgo_CellsBuilder
  (const Handle(NCollection_BaseAllocator)& theAllocator)
:
  BOPAlgo_Builder(theAllocator),
  myType(TopAbs_SHAPE),
  myIndex(100, myAllocator),
  myMaterials(100, myAllocator),
  myShapeMaterial(100, myAllocator),
  myMapModified(100, myAllocator)
{
}

//=======================================================================
//function : ~
//purpose  : 
//=======================================================================
BOPAlgo_CellsBuilder::~BOPAlgo_CellsBuilder()
{
  Clear();
}


//=======================================================================
//function : Clear
//purpose  : 
//=======================================================================
void BOPAlgo_CellsBuilder::Clear()
{
  BOPAlgo_Builder::Clear();
  myIndex.Clear();
  myMaterials.Clear();
  myShapeMaterial.Clear();
  myMapModified.Clear();
}

//=======================================================================
//function : GetAllParts
//purpose  : 
//=======================================================================
const TopoDS_Shape& BOPAlgo_CellsBuilder::GetAllParts() const
{
  return myAllParts;
}

//=======================================================================
//function : Prepare
//purpose  : 
//=======================================================================
void BOPAlgo_CellsBuilder::Prepare()
{
  BOPAlgo_Builder::Prepare();
  //
  myFlagHistory=Standard_False;
}

//=======================================================================
// function: CheckData
// purpose: 
//=======================================================================
void BOPAlgo_CellsBuilder::CheckData()
{
  BOPAlgo_Builder::CheckData();
  if (myErrorStatus) {
    return;
  }
  //
  // additional check for the arguments to be of the same dimension.
  Standard_Integer aDim1, aDimi;
  BOPCol_ListIteratorOfListOfShape aIt;
  //
  aIt.Initialize(myArguments);
  const TopoDS_Shape& aS1 = aIt.Value();
  aDim1 = BOPTools_AlgoTools::Dimension(aS1);
  //
  for (aIt.Next(); aIt.More(); aIt.Next()) {
    const TopoDS_Shape& aSi = aIt.Value();
    aDimi = BOPTools_AlgoTools::Dimension(aSi);
    //
    if (aDim1 != aDimi) {
      myErrorStatus = 201; // non-homogenous arguments
      break;
    }
  }
}

//=======================================================================
//function : PerformInternal1
//purpose  : 
//=======================================================================
void BOPAlgo_CellsBuilder::PerformInternal1(const BOPAlgo_PaveFiller& theFiller)
{
  BOPAlgo_Builder::PerformInternal1(theFiller);
  //
  if (myErrorStatus) {
    return;
  }
  //
  // save the splits to <myAllParts>
  TakeAllParts();
  //
  // index all the parts to its origins
  IndexParts();
  //
  // and nullify <myShape> for building the result;
  RemoveAllFromResult();
  //
  myFlagHistory = Standard_True;
}

//=======================================================================
//function : TakeAllParts
//purpose  : 
//=======================================================================
void BOPAlgo_CellsBuilder::TakeAllParts()
{
  Standard_Integer aDim;
  TopoDS_Compound aC;
  BRep_Builder aBB;
  //
  aDim = BOPTools_AlgoTools::Dimension(myArguments.First());
  myType = TypeToExplore(aDim);
  //
  aBB.MakeCompound(aC);
  TopExp_Explorer aExp(myShape, myType);
  for (; aExp.More(); aExp.Next()) {
    const TopoDS_Shape& aS = aExp.Current();
    aBB.Add(aC, aS);
  }
  myAllParts = aC;
}

//=======================================================================
//function : IndexParts
//purpose  : 
//=======================================================================
void BOPAlgo_CellsBuilder::IndexParts()
{
  BOPCol_ListIteratorOfListOfShape aIt, aItIm;
  TopExp_Explorer aExp;
  //
  aIt.Initialize(myArguments);
  for (; aIt.More(); aIt.Next()) {
    const TopoDS_Shape& aS = aIt.Value();
    //
    aExp.Init(aS, myType);
    for (; aExp.More(); aExp.Next()) {
      const TopoDS_Shape& aST = aExp.Current();
      const BOPCol_ListOfShape* pLSIm = myImages.Seek(aST);
      if (!pLSIm) {
        BOPCol_ListOfShape* pLS = myIndex.ChangeSeek(aST);
        if (!pLS) {
          pLS = &myIndex(myIndex.Add(aST, BOPCol_ListOfShape()));
        }
        pLS ->Append(aS);
        continue;
      }
      //
      aItIm.Initialize(*pLSIm);
      for (; aItIm.More(); aItIm.Next()) {
        const TopoDS_Shape& aSTIm = aItIm.Value();
        //
        BOPCol_ListOfShape* pLS = myIndex.ChangeSeek(aSTIm);
        if (!pLS) {
          pLS = &myIndex(myIndex.Add(aSTIm, BOPCol_ListOfShape()));
        }
        pLS ->Append(aS);
      } // for (; aItIm.More(); aItIm.Next()) {
    } // for (; aExp.More(); aExp.Next()) {
  } // for (; aIt.More(); aIt.Next()) {
}

//=======================================================================
//function : AddToResult
//purpose  : 
//=======================================================================
void BOPAlgo_CellsBuilder::AddToResult(const BOPCol_ListOfShape& theLSToTake,
                                       const BOPCol_ListOfShape& theLSToAvoid,
                                       const Standard_Integer theMaterial,
                                       const Standard_Boolean theUpdate)
{
  // find parts
  BOPCol_ListOfShape aParts;
  FindParts(theLSToTake, theLSToAvoid, aParts);
  if (aParts.IsEmpty()) {
    return;
  }
  //
  Standard_Boolean bChanged;
  BRep_Builder aBB;
  BOPCol_MapOfShape aResParts;
  TopExp_Explorer aExp;
  BOPCol_ListIteratorOfListOfShape aIt;
  //
  bChanged = Standard_False;
  // collect result parts to avoid multiple adding of the same parts
  aExp.Init(myShape, myType);
  for (; aExp.More(); aExp.Next()) {
    const TopoDS_Shape& aPart = aExp.Current();
    aResParts.Add(aPart);
  }
  // add parts to result 
  aIt.Initialize(aParts);
  for (; aIt.More(); aIt.Next()) {
    const TopoDS_Shape& aPart = aIt.Value();
    if (aResParts.Add(aPart)) {
      aBB.Add(myShape, aPart);
      bChanged = Standard_True;
    }
  }
  //
  // update the material
  if (theMaterial != 0) {
    BOPCol_ListOfShape aLSP;
    aIt.Initialize(aParts);
    for (; aIt.More(); aIt.Next()) {
      const TopoDS_Shape& aPart = aIt.Value();
      if (!myShapeMaterial.IsBound(aPart)) {
        myShapeMaterial.Bind(aPart, theMaterial);
        aLSP.Append(aPart);
      }
    } // for (; aIt.More(); aIt.Next()) {
    //
    if (aLSP.Extent()) {
      if (myMaterials.IsBound(theMaterial)) {
        BOPCol_ListOfShape& aLS = myMaterials.ChangeFind(theMaterial);
        aLS.Append(aLSP);
      } // if (myMaterials.IsBound(theMaterial)) {
      else {
        myMaterials.Bind(theMaterial, aLSP);
      }
    } // if (aLSP.Extent()) {
  } // if (theMaterial != 0) {
  //
  if (!theUpdate) {
    if (bChanged) {
      PrepareHistory();
    }
  }
  else {
    RemoveInternalBoundaries();
  }
}

//=======================================================================
//function : AddAllToResult
//purpose  : 
//=======================================================================
void BOPAlgo_CellsBuilder::AddAllToResult(const Standard_Integer theMaterial,
                                          const Standard_Boolean theUpdate)
{
  TopoDS_Compound aResult;
  BRep_Builder aBB;
  BOPCol_ListOfShape aLSM;
  //
  aBB.MakeCompound(aResult);
  myShapeMaterial.Clear();
  myMaterials.Clear();
  myMapModified.Clear();
  //
  TopoDS_Iterator aIt(myAllParts);
  for (; aIt.More(); aIt.Next()) {
    const TopoDS_Shape& aPart = aIt.Value();
    aBB.Add(aResult, aPart);
    //
    if (theMaterial != 0) {
      myShapeMaterial.Bind(aPart, theMaterial);
      aLSM.Append(aPart);
    }
  }
  //
  myShape = aResult;
  //
  if (theMaterial != 0) {
    myMaterials.Bind(theMaterial, aLSM);
  }
  //
  if (!theUpdate) {
    PrepareHistory();
  }
  else {
    RemoveInternalBoundaries();
  }
}

//=======================================================================
//function : RemoveFromResult
//purpose  : 
//=======================================================================
void BOPAlgo_CellsBuilder::RemoveFromResult(const BOPCol_ListOfShape& theLSToTake,
                                            const BOPCol_ListOfShape& theLSToAvoid)
{
  // find parts
  BOPCol_ListOfShape aParts;
  FindParts(theLSToTake, theLSToAvoid, aParts);
  if (aParts.IsEmpty()) {
    return;
  }
  //
  BOPCol_MapOfShape aPartsToRemove;
  BOPCol_ListIteratorOfListOfShape aItP, aItM;
  //
  // collect parts into the map and remove parts from materials
  aItP.Initialize(aParts);
  for (; aItP.More(); aItP.Next()) {
    const TopoDS_Shape& aPart = aItP.Value();
    aPartsToRemove.Add(aPart);
    //
    if (myShapeMaterial.IsBound(aPart)) {
      Standard_Integer iMaterial = myShapeMaterial.Find(aPart);
      if (!myMaterials.IsBound(iMaterial)) {
        myShapeMaterial.UnBind(aPart);
        continue;
      }
      //
      BOPCol_ListOfShape& aLSM = myMaterials.ChangeFind(iMaterial);
      //
      aItM.Initialize(aLSM);
      for (; aItM.More(); aItM.Next()) {
        const TopoDS_Shape& aSM = aItM.Value();
        if (aSM.IsSame(aPart)) {
          aLSM.Remove(aItM);
          break;
        } // if (aSM.IsSame(aPart)) {
      } // for (; aItM.More(); aItM.Next()) {
      //
      myShapeMaterial.UnBind(aPart);
    } // if (myShapeMaterial.IsBound(aPart)) {
  } // for (; aItP.More(); aItP.Next()) {
  //
  Standard_Boolean bChanged;
  TopoDS_Compound aResult;
  BRep_Builder aBB;
  TopoDS_Iterator aIt1, aIt2;
  TopAbs_ShapeEnum aType;
  //
  aBB.MakeCompound(aResult);
  bChanged = Standard_False;
  //
  aIt1.Initialize(myShape);
  for (; aIt1.More(); aIt1.Next()) {
    const TopoDS_Shape& aS = aIt1.Value();
    aType = aS.ShapeType();
    //
    if (aType == myType) {
      // basic element
      if (aPartsToRemove.Contains(aS)) {
        bChanged = Standard_True;
        continue;
      }
      aBB.Add(aResult, aS);
    } // if (aType == myType) {
    else {
      // container
      Standard_Boolean bNotEmpty;
      TopoDS_Compound aSC;
      aBB.MakeCompound(aSC);
      //
      bNotEmpty = Standard_False;
      aIt2.Initialize(aS);
      for (; aIt2.More(); aIt2.Next()) {
        const TopoDS_Shape& aSS = aIt2.Value();
        if (aPartsToRemove.Contains(aSS)) {
          bChanged = Standard_True;
          continue;
        }
        aBB.Add(aSC, aSS);
        bNotEmpty = Standard_True;
      } // for (; aIt2.More(); aIt2.Next()) {
      //
      if (bNotEmpty) {
        MakeTypedContainers(aSC, myType, aResult);
      }
    } // else {
  } // for (; aIt1.More(); aIt1.Next()) {
  //
  if (bChanged) {
    myShape = aResult;
    //
    PrepareHistory();
  }
}

//=======================================================================
//function : RemoveAllFromResult
//purpose  : 
//=======================================================================
void BOPAlgo_CellsBuilder::RemoveAllFromResult()
{
  // empty compound
  TopoDS_Compound aC;
  BRep_Builder aBB;
  //
  aBB.MakeCompound(aC);
  myShape = aC;
  //
  myMaterials.Clear();
  myShapeMaterial.Clear();
  myMapModified.Clear();
  //
  PrepareHistory();
}

//=======================================================================
//function : RemoveInternalBoundaries
//purpose  : 
//=======================================================================
void BOPAlgo_CellsBuilder::RemoveInternalBoundaries()
{
  if (myMaterials.IsEmpty()) {
    return;
  }
  //
  Standard_Integer iMaterial, iErr;
  TopoDS_Compound aResult;
  BRep_Builder aBB;
  TopExp_Explorer aExp;
  BOPCol_ListIteratorOfListOfShape aItS;
  BOPCol_DataMapIteratorOfDataMapOfIntegerListOfShape aItM;
  //
  aBB.MakeCompound(aResult);
  aExp.Init(myShape, myType);
  for (; aExp.More(); aExp.Next()) {
    const TopoDS_Shape& aS = aExp.Current();
    if (!myShapeMaterial.IsBound(aS)) {
      aBB.Add(aResult, aS);
    }
  }
  //
  aItM.Initialize(myMaterials);
  for (; aItM.More(); aItM.Next()) {
    iMaterial = aItM.Key();
    //
    BOPCol_ListOfShape aLSNew;
    BOPCol_ListOfShape& aLS = myMaterials(iMaterial);
    iErr = RemoveInternals(aLS, aLSNew);
    if (iErr || aLSNew.IsEmpty()) {
      myErrorStatus = 202; // unable to remove internal boundaries
      return;
    }
    //
    // update materials maps and add new shapes to result
    aLS.Assign(aLSNew);
    aItS.Initialize(aLSNew);
    for (; aItS.More(); aItS.Next()) {
      const TopoDS_Shape& aS = aItS.Value();
      aBB.Add(aResult, aS);
      if (!myShapeMaterial.IsBound(aS)) {
        myShapeMaterial.Bind(aS, iMaterial);
      }
    }
  }
  //
  myShape = aResult;
  //
  PrepareHistory();
}

//=======================================================================
//function : FindPart
//purpose  : 
//=======================================================================
void BOPAlgo_CellsBuilder::FindParts(const BOPCol_ListOfShape& theLSToTake,
                                     const BOPCol_ListOfShape& theLSToAvoid,
                                     BOPCol_ListOfShape& theParts)
{
  if (theLSToTake.IsEmpty()) {
    return;
  }
  //
  Standard_Boolean bFound;
  Standard_Integer aNbS;
  BOPCol_ListIteratorOfListOfShape aItIm, aItArgs;
  BOPCol_MapOfShape aMSToTake, aMSToAvoid, aMS;
  TopExp_Explorer aExp;
  //
  aItArgs.Initialize(theLSToAvoid);
  for (; aItArgs.More(); aItArgs.Next()) {
    const TopoDS_Shape& aS = aItArgs.Value();
    aMSToAvoid.Add(aS);
  }
  //
  aItArgs.Initialize(theLSToTake);
  for (; aItArgs.More(); aItArgs.Next()) {
    const TopoDS_Shape& aS = aItArgs.Value();
    aMSToTake.Add(aS);
  }
  //
  aNbS = aMSToTake.Extent();
  //
  const TopoDS_Shape& aSToTake = theLSToTake.First();
  aExp.Init(aSToTake, myType);
  for (; aExp.More(); aExp.Next()) {
    const TopoDS_Shape& aST = aExp.Current();
    BOPCol_ListOfShape aLSTIm;
    if (!myImages.IsBound(aST)) {
      aLSTIm.Append(aST);
    } else {
      aLSTIm = myImages.Find(aST);
    }
    //
    aItIm.Initialize(aLSTIm);
    for (; aItIm.More(); aItIm.Next()) {
      const TopoDS_Shape& aPart = aItIm.Value();
      //
      if (!myIndex.Contains(aPart)) {
        continue;
      }
      //
      const BOPCol_ListOfShape& aLS = myIndex.FindFromKey(aPart);
      if (aLS.Extent() < aNbS) {
        continue;
      }
      //
      aMS.Clear();
      aItArgs.Initialize(aLS);
      for (bFound = Standard_True; aItArgs.More() && bFound; aItArgs.Next()) {
        const TopoDS_Shape& aS = aItArgs.Value();
        bFound = !aMSToAvoid.Contains(aS);
        aMS.Add(aS);
      }
      //
      if (!bFound) {
        continue;
      }
      //
      aItArgs.Initialize(theLSToTake);
      for (; aItArgs.More() && bFound; aItArgs.Next()) {
        const TopoDS_Shape& aS = aItArgs.Value();
        bFound = aMS.Contains(aS);
      }
      //
      if (bFound) {
        theParts.Append(aPart);
      } // if (bFound) {
    } // for (; aItIm.More(); aItIm.Next()) {
  } // for (; aExp.More(); aExp.Next()) {
}

//=======================================================================
//function : MakeContainers
//purpose  : 
//=======================================================================
void BOPAlgo_CellsBuilder::MakeContainers()
{
  TopoDS_Compound aResult;
  BRep_Builder aBB;
  //
  aBB.MakeCompound(aResult);
  MakeTypedContainers(myShape, myType, aResult);
  myShape = aResult;
}

//=======================================================================
//function : RemoveInternals
//purpose  : 
//=======================================================================
Standard_Integer BOPAlgo_CellsBuilder::RemoveInternals(const BOPCol_ListOfShape& theLS,
                                                       BOPCol_ListOfShape& theLSNew)
{
  Standard_Integer iErr = 0;
  if (theLS.Extent() < 2) {
    theLSNew = theLS;
    return iErr;
  }
  //  
  TopExp_Explorer aExp;
  //
  TopAbs_ShapeEnum aType = theLS.First().ShapeType();
  //
  if (aType == TopAbs_EDGE ||
      aType == TopAbs_FACE) {
    //
    // make container
    BRep_Builder aBB;
    TopoDS_Shape aShape;
    //
    BOPTools_AlgoTools::MakeContainer
      ((aType == TopAbs_FACE) ? TopAbs_SHELL : TopAbs_WIRE, aShape);
    //
    for (BOPCol_ListIteratorOfListOfShape aIt(theLS); aIt.More(); aIt.Next()) {
      const TopoDS_Shape& aS = aIt.Value();
      aBB.Add(aShape, aS);
    }
    //
    // Unify same domain
    Standard_Boolean bFaces, bEdges;
    //
    bFaces = (aType == TopAbs_FACE);
    bEdges = (aType == TopAbs_EDGE);
    //
    ShapeUpgrade_UnifySameDomain anUnify (aShape, bEdges, bFaces);
    anUnify.Build();
    const TopoDS_Shape& aSNew = anUnify.Shape();
    //
    aExp.Init(aSNew, aType);
    for (; aExp.More(); aExp.Next()) {
      const TopoDS_Shape& aSn = aExp.Current();
      theLSNew.Append(aSn);
    }
    //
    // fill map of generated shapes
    BOPCol_IndexedMapOfShape aMG;
    Standard_Integer i, aNb;
    //
    BOPTools::MapShapes(aShape, TopAbs_VERTEX, aMG);
    BOPTools::MapShapes(aShape, TopAbs_EDGE,   aMG);
    BOPTools::MapShapes(aShape, TopAbs_FACE,   aMG);
    //
    aNb = aMG.Extent();
    for (i = 1; i <= aNb; ++i) {
      const TopoDS_Shape& aSS = aMG(i);
      const TopoDS_Shape& aSGen = anUnify.Generated(aSS);
      if (!aSGen.IsNull() && !aSS.IsSame(aSGen)) {
        myMapModified.Bind(aSS, aSGen);
      }
    }
  }
  else if (aType == TopAbs_SOLID) {
    BRep_Builder aBB;
    TopoDS_Compound aSolids;
    aBB.MakeCompound(aSolids);
    //
    BOPCol_ListIteratorOfListOfShape aItLS(theLS);
    for (; aItLS.More(); aItLS.Next()) {
      const TopoDS_Shape& aSol = aItLS.Value();
      aBB.Add(aSolids, aSol);
    }
    //
    // Make connexity blocks of solids to create from each isolated block one solid.
    // It will allow attaching internal entities of the solids to new solid.
    BOPCol_ListOfShape aLCB;
    BOPTools_AlgoTools::MakeConnexityBlocks(aSolids, TopAbs_FACE, TopAbs_SOLID, aLCB);
    //
    // for each block remove internal faces
    BOPCol_ListIteratorOfListOfShape aItLCB(aLCB);
    for (; aItLCB.More(); aItLCB.Next()) {
      const TopoDS_Shape& aCB = aItLCB.Value();
      //
      // Map faces and solids to find boundary faces that can be removed
      BOPCol_IndexedDataMapOfShapeListOfShape aDMFS;
      // internal entities
      BOPCol_ListOfShape aLSInt;
      //
      TopoDS_Iterator aItS(aCB);
      for (; aItS.More(); aItS.Next()) {
        const TopoDS_Shape& aSol = aItS.Value();
        //
        TopoDS_Iterator aItIS(aSol);
        for (; aItIS.More(); aItIS.Next()) {
          const TopoDS_Shape& aSI = aItIS.Value();
          if (aSI.Orientation() == TopAbs_INTERNAL) {
            aLSInt.Append(aSI);
          }
          else {
            TopoDS_Iterator aItF(aSI);
            for (; aItF.More(); aItF.Next()) {
              const TopoDS_Shape& aF = aItF.Value();
              BOPCol_ListOfShape *pLSols = aDMFS.ChangeSeek(aF);
              if (!pLSols) {
                pLSols = &aDMFS(aDMFS.Add(aF, BOPCol_ListOfShape()));
              }
              pLSols->Append(aSol);
            }
          }
        }
      }
      //
      // to build unified solid, select only faces attached to only one solid
      BOPCol_ListOfShape aLFUnique;
      Standard_Integer i, aNb = aDMFS.Extent();
      for (i = 1; i <= aNb; ++i) {
        if (aDMFS(i).Extent() == 1) {
          aLFUnique.Append(aDMFS.FindKey(i));
        }
      }
      //
      if (aNb == aLFUnique.Extent()) {
        // no faces to remove
        aItS.Initialize(aCB);
        for (; aItS.More(); aItS.Next()) {
          theLSNew.Append(aItS.Value());
        }
        continue;
      }
      //
      // build new solid
      BOPAlgo_BuilderSolid aBS;
      aBS.SetShapes(aLFUnique);
      aBS.Perform();
      //
      iErr = aBS.ErrorStatus();
      if (!iErr && aBS.Areas().Extent() != 1)
        iErr = 1;
      if (iErr)
        return iErr;

      TopoDS_Solid& aSNew = *(TopoDS_Solid*)&aBS.Areas().First();
      //
      // put all internal parts into new solid
      aSNew.Free(Standard_True);
      BOPCol_ListIteratorOfListOfShape aItLSI(aLSInt);
      for (; aItLSI.More(); aItLSI.Next()) {
        aBB.Add(aSNew, aItLSI.Value());
      }
      aSNew.Free(Standard_False);
      //
      theLSNew.Append(aSNew);

      // Save information about the fuse of the solids into a history map
      aItS.Initialize(aCB);
      for (; aItS.More(); aItS.Next())
        myMapModified.Bind(aItS.Value(), aSNew);
    }
  }
  //
  return iErr;
}

//=======================================================================
//function : PrepareHistory
//purpose  : 
//=======================================================================
void BOPAlgo_CellsBuilder::PrepareHistory()
{
  if (!myFlagHistory)
  {
    // Clearing
    BOPAlgo_BuilderShape::PrepareHistory();
    return;
  }

  // Clearing from previous operations
  BOPAlgo_BuilderShape::PrepareHistory();
  myFlagHistory = Standard_True;

  if (myShape.IsNull() ||
      BOPTools_AlgoTools3D::IsEmptyShape(myShape))
  {
    // The result shape is a null shape or empty shape,
    // thus, no modified, no generated, all deleted
    myHasModified = Standard_False;
    myHasGenerated = Standard_False;
    myHasDeleted = Standard_True;
    return;
  }

  // Map the result shape
  BOPTools::MapShapes(myShape, myMapShape);

  // Among all input shapes find those that have any trace in the result
  // and save them into myImagesResult map with connection to parts
  // kept in the result shape.
  // Also, set the proper values to the history flags:
  // - myHasDeleted for Deleted shapes;
  // - myHasModified for Modified shapes;
  // - myHasGenerated for Generated shapes.
  Standard_Integer aNbS = myDS->NbSourceShapes();
  for (Standard_Integer i = 0; i < aNbS; ++i)
  {
    const TopoDS_Shape& aS = myDS->Shape(i);

    // History information is only available for the shapes of type
    // VERTEX, EDGE, FACE and SOLID. Skip all shapes of different type.
    TopAbs_ShapeEnum aType = aS.ShapeType();
    if (!(aType == TopAbs_VERTEX ||
          aType == TopAbs_EDGE   ||
          aType == TopAbs_FACE   ||
          aType == TopAbs_SOLID))
      continue;

    // Check if the shape has any splits
    const BOPCol_ListOfShape* pLSp = myImages.Seek(aS);
    if (!pLSp)
    {
      // No splits from GF builder.
      // Check if the shape has been locally modified,
      // i.e. has participated in the Removal of internal boundaries
      // and have been unified with other shapes
      const TopoDS_Shape* pSU = myMapModified.Seek(aS);
      // Check if the result contains the shape or its unified image
      const TopoDS_Shape* pSR = !pSU ? &aS : pSU;
      if (myMapShape.Contains(*pSR))
      {
        // Link the shape to itself or to its unified image
        myImagesResult(myImagesResult.Add(aS, BOPCol_ListOfShape())).Append(*pSR);
        if (pSU)
          // Modified element is found
          myHasModified = Standard_True;
      }
      else
      {
        // No trace of the shape in the result -> Deleted element is found
        myHasDeleted = Standard_True;
      }
    }
    else
    {
      // Find all splits of the shape which are kept in the result
      BOPCol_ListOfShape *pLSpKept = NULL;
      BOPCol_ListIteratorOfListOfShape aIt(*pLSp);
      for (; aIt.More(); aIt.Next())
      {
        TopoDS_Shape aSp = aIt.Value();

        // Check SD shapes
        const TopoDS_Shape* pSpSD = myShapesSD.Seek(aSp);
        if (pSpSD)
          aSp = *pSpSD;

        // Check local modification of the shape (unification with other shapes)
        const TopoDS_Shape* pSU = myMapModified.Seek(aSp);
        if (pSU)
          aSp = *pSU;

        // Check if the result shape contains the split
        if (myMapShape.Contains(aSp))
        {
          if (!pLSpKept)
            pLSpKept = &myImagesResult(myImagesResult.Add(aS, BOPCol_ListOfShape()));

          // Link the shape to the split
          pLSpKept->Append(aSp);
        }
      }

      if (pLSpKept)
        // Modified element is found
        myHasModified = Standard_True;
      else
        // Deleted element is found
        myHasDeleted = Standard_True;
    }

    // Try to find generated shapes
    if (!myHasGenerated)
    {
      // Temporarily set the HasGenerated flag to look for the shapes generated from aS
      myHasGenerated = Standard_True;
      myHasGenerated = (Generated(aS).Extent() > 0);
    }
  }
}

//=======================================================================
//function : MakeTypedContainers
//purpose  : 
//=======================================================================
void MakeTypedContainers(const TopoDS_Shape& theSC,
                         const TopAbs_ShapeEnum aType,
                         TopoDS_Shape& theResult)
{
  TopAbs_ShapeEnum aContainerType, aConnexityType, aPartType;
  //
  aPartType = aType;
  switch (aPartType) {
    case TopAbs_EDGE: {
      aContainerType = TopAbs_WIRE;
      aConnexityType = TopAbs_VERTEX;
      break;
    }
    case TopAbs_FACE: {
      aContainerType = TopAbs_SHELL;
      aConnexityType = TopAbs_EDGE;
      break;
    }
    case TopAbs_SOLID: {
      aContainerType = TopAbs_COMPSOLID;
      aConnexityType = TopAbs_FACE;
      break;
    }
    default:
      return;
  }
  //
  BOPCol_ListOfShape aLCB;
  BOPTools_AlgoTools::MakeConnexityBlocks(theSC, aConnexityType, aPartType, aLCB);
  if (aLCB.IsEmpty()) {
    return;
  }
  //
  BRep_Builder aBB;
  TopExp_Explorer aExp;
  BOPCol_ListIteratorOfListOfShape aItCB;
  //
  aItCB.Initialize(aLCB);
  for (; aItCB.More(); aItCB.Next()) {
    TopoDS_Shape aRCB;
    BOPTools_AlgoTools::MakeContainer(aContainerType, aRCB);
    //
    const TopoDS_Shape& aCB = aItCB.Value();
    aExp.Init(aCB, aPartType);
    for (; aExp.More(); aExp.Next()) {
      const TopoDS_Shape& aCBS = aExp.Current();
      aBB.Add(aRCB, aCBS);
    }
    //
    if (aContainerType == TopAbs_SHELL) {
      BOPTools_AlgoTools::OrientFacesOnShell(aRCB);
    }
    //
    aBB.Add(theResult, aRCB);
  }
}

//=======================================================================
//function : TypeToExplore
//purpose  : 
//=======================================================================
TopAbs_ShapeEnum TypeToExplore(const Standard_Integer theDim)
{
  TopAbs_ShapeEnum aRet;
  //
  switch(theDim) {
  case 0:
    aRet=TopAbs_VERTEX;
    break;
  case 1:
    aRet=TopAbs_EDGE;
    break;
  case 2:
    aRet=TopAbs_FACE;
    break;
  case 3:
    aRet=TopAbs_SOLID;
    break;
  default:
    aRet=TopAbs_SHAPE;
    break;
  }
  return aRet;
}
